﻿using System;
using System.Collections.Generic;

namespace PoCFluentInterfaceOnDomains
{
    public class Category 
    {
        private readonly IList<CatalogItem> _items;
        public virtual String Name { get; private set; }

        public IList<CatalogItem> Items
        {
            get { return _items; }
        }
        public Category()
        {
            _items = new List<CatalogItem>();
        }
        public void Categorize(CatalogItem item)
        {
            Items.Add(item);
        }
        public void Categorize(Action<CatalogItemBuilder> buildUsing)
        {
            var catalogItemBuilder = new CatalogItemBuilder();
            buildUsing(catalogItemBuilder);
            var item = catalogItemBuilder.Build();
            Categorize(item);
        }
    }

    public class CatalogItem
    {
        public string Name { get; set; }
        public Manufacturer Manufacturer { get; set; }
        public Product Product { get; set; }
        public CatalogItem(string name, Manufacturer manufacturer, Product product)
        {
            Name = name;
            Manufacturer = manufacturer;
            Product = product;
        }
    }

    public class Manufacturer 
    {
        public string Name { get; set; }
        public Manufacturer(string name)
        {
            Name = name;
        }
    }

    public class Product 
    {
        public string Number { get; set; }

        public Product(string number)
        {
            Number = number;
        }
    }

    public interface ICatalogItemBuilder
    {
        CatalogItem Build();
    }
  
    public abstract class DomainObjectBuilder<T>
    {
        public abstract T Build();
    }

    public class CatalogItemBuilder : DomainObjectBuilder<CatalogItem>, ICatalogItemBuilder
    {
        protected Manufacturer Manufacturer { get; set; }
        protected String Name { get; set; }
        protected Product Product { get; set; }

        public CatalogItemBuilder Named(String name)
        {
            Name = name;
            return this;
        }

        public CatalogItemBuilder ManufacturedBy(Action<ManufacturerBuilder> buildUsing)
        {
            var manufacturerBuilder = new ManufacturerBuilder();
            buildUsing(manufacturerBuilder);
            Manufacturer = manufacturerBuilder.Build();
            return this;
        }
        public CatalogItemBuilder ForProduct(Action<ProductBuilder> buildUsing)
        {
            var productBuilder = new ProductBuilder();
            buildUsing(productBuilder);
            Product = productBuilder.Build();
            return this;
        }
        public override CatalogItem Build()
        {
            return new CatalogItem(Name, Manufacturer, Product);
        }

        public static implicit operator CatalogItem(CatalogItemBuilder builder)
        {
            return builder.Build();
        }

    }

    public class ManufacturerBuilder : DomainObjectBuilder<Manufacturer>
    {
        public override Manufacturer Build()
        {
            return new Manufacturer(Name);
        }
        public ManufacturerBuilder Named(string name)
        {
            Name = name;
            return this;
        }
        protected string Name { get; set; }
    }
    
    public class ProductBuilder : DomainObjectBuilder<Product>
    {

        public override Product Build()
        {
            return new Product(Number);
        }

        public ProductBuilder Numbered(string number)
        {
            Number = number;
            return this;
        }
        public string Number { get; set; }
    }

    public static class CatalogCategoryExtensions
    {
        public static void Categorize(this Category catalogCategory,Action<CatalogItemBuilder> buildUsing)
        {
            var catalogItemBuilder = new CatalogItemBuilder();
            buildUsing(catalogItemBuilder);
            catalogCategory.Categorize(catalogItemBuilder);
        }

    }


}
